#include <stdio.h>
#include <stdlib.h>
#include "../include/utilities.h"
#include "../include/meta_explorer.h"

#define MAX_FILE_NAME_LENGTH 29
#define MAX_LINES 15
#define MAX_ROWS 3

int totalSwaps = 0;
int success = 0;

void displayResults();

void applyPatch(FileBlock** filesToPatch, int filesToPatchCount,MetaFileInfo* metaFileInfo)
{
    if (!backupExists())
    {
        createBackup();
    }
    FILE* metaFile = openFile("pad00000.meta","rb+");
    MemFileInt* metaFileMem = copyFileToMemoryAsInt(metaFile);
    int i = 0;

    printf("\n\n======== Patching Meta File ========");
    for (i = 0; i < filesToPatchCount; i++)
    {
        printf("\n(%d/%d): %s",i + 1, filesToPatchCount, filesToPatch[i]->fileName);

        long hashLocation = locateHash(filesToPatch[i]->hash,metaFileMem,metaFileInfo);

        if (hashLocation != NOT_FOUND)
        {
            metaFileMem->block[hashLocation+1] = RANDOM_FOLDER_NUM;
            metaFileMem->block[hashLocation+2] = RANDOM_FILE_NUM;
            printColor(" Success", GREEN);
        }
        else
        {
            printColor(" Failed!", RED);
        }

    }

    rewind(metaFile);
    for (i = 0; i < metaFileMem->size; i++)
    {
        fwrite(&metaFileMem->block[i],sizeof(int),1,metaFile);
    }

    fclose(metaFile);
    printf("\n");
}





char* mainHeader()
{
    char* mainHeader[] = {
    "******************************************************\n",
    "*                                                    *\n",
    "*          Black Desert Online Armor Swapping        *\n",
    "*                                                    *\n",
    "******************************************************\n",
    };
    int i = 0;
    int headerLen = 0;
    for (i = 0; i < 5; i++)
    {
        headerLen += strlen(mainHeader[i]);
    }
    char* headerString = (char*)malloc(headerLen + 1);
    strcpy(headerString,mainHeader[0]);
    for (i = 1; i < 5; i++)
    {
        strcat(headerString,mainHeader[i]);
    }
    return headerString;
}

void swapArmors(char* armor1, char* armor2, FileBlock* fileBlocks, MetaFileInfo* metaFileInfo)
{
    int i = 0;
    int partsToPatchCount = 0;
    FileBlock* partsToPatch1[25];
    FileBlock* partsToPatch2[25];


    partsToPatch1[0] = binarySearchFileBlock("common_decals_road_04_dec.dds",fileBlocks, metaFileInfo->fileBlocksCount);

    //copyFileTo("patcher_resources\\texture\\blank_dxt1.dds",concatenate(bdoRootFolder(),"object\\texture\\common_decals_road_04_dec.dds"));

    partsToPatchCount = 1;

    if(contains(armor1,"ub") && contains(armor2,"ub"))
    {
        for (i = 0; i < metaFileInfo->fileBlocksCount; i++)
        {
           if (fileHasExt(fileBlocks[i].fileName,"pac")
                && (contains(fileBlocks[i].folderName,"armor"))
                && (strcmpi(getClassPrefix(fileBlocks[i].fileName),getClassPrefix(armor1)) == 0)
                && (strcmpi(getTwoDigits(fileBlocks[i].fileName),getTwoDigits(armor1)) == 0)
                && (strcmpi(getFourDigits(fileBlocks[i].fileName),getFourDigits(armor1)) == 0))
             {
                   if (!contains(fileBlocks[i].fileName,"uw"))
                    {
                        FileBlock* partArmor1 = &fileBlocks[i];

                        char* part2Name = (char*) malloc(strlen(partArmor1->fileName));
                        strcpy(part2Name,partArmor1->fileName);

                        strreplace(&part2Name,getClassPrefix(part2Name),getClassPrefix(armor2));
                        strreplace(&part2Name,getTwoDigits(part2Name),getTwoDigits(armor2));
                        strreplace(&part2Name,getFourDigits(part2Name),getFourDigits(armor2));

                        FileBlock* partArmor2 = binarySearchFileBlock(part2Name,fileBlocks, metaFileInfo->fileBlocksCount);

                        if (contains(armor2,"00_ub_0034")  && ((contains(part2Name,"sho") || contains(part2Name,"cloak") || contains(part2Name,"lb"))) // Kibelius fix
                        || (contains(armor2,"10_ub_0006")  && (armor2[2] != 'm') && (armor2[2] != 'M')))
                        {
                            // Skip
                            if (((contains(part2Name,"sho")  && contains(armor2,"_0034_01"))) ||  !contains(armor2,"_0034"))
                            {
                                partArmor2 = binarySearchFileBlock("t0072_pumpkin_ground_0001.pac",fileBlocks, metaFileInfo->fileBlocksCount);
                            }
                        }

                        if (partArmor2 == NULL)
                        {
                            partArmor2 = binarySearchFileBlock("t0072_pumpkin_ground_0001.pac",fileBlocks, metaFileInfo->fileBlocksCount);
                        }


                        if (partArmor2 != NULL)
                        {
                            partsToPatch1[partsToPatchCount] = partArmor1;
                            partsToPatch2[partsToPatchCount] = partArmor2;
                            //printf("\n--- %s -> %s", partsToPatch2[partsToPatchCount]->fileName, partsToPatch1[partsToPatchCount]->fileName);
                            partsToPatchCount++;
                        }
                    }
             }
        }
    }
    else
    {
        partsToPatchCount = 1;
        partsToPatch1[0] = binarySearchFileBlock(armor1,fileBlocks, metaFileInfo->fileBlocksCount);
        partsToPatch2[0] = binarySearchFileBlock(armor2,fileBlocks, metaFileInfo->fileBlocksCount);
    }

    for (i = 1; i < partsToPatchCount; i++)
    {
        printf("\n%s -> %s", partsToPatch2[i]->fileName,partsToPatch1[i]->fileName);

        char* armor1FullPath = concatenate(bdoRootFolder(), partsToPatch1[i]->folderName);
        createPath(armor1FullPath);

        armor1FullPath = concatenate(armor1FullPath, partsToPatch1[i]->fileName);

        char* armor2FullPath = concatenate(bdoRootFolder(), partsToPatch2[i]->folderName);


        armor2FullPath = concatenate(armor2FullPath, partsToPatch2[i]->fileName);


        if( DEBUG)
        {
            extractFileCustomPath(partsToPatch2[i],concatenate("../",partsToPatch2[i]->folderName),1);
        }
        else
        {
            extractFileCustomPath(partsToPatch2[i],concatenate("../",partsToPatch2[i]->folderName),0);
        }


        charReplace(armor2FullPath,'/','\\');


        if (fileExists(armor1FullPath))
        {
            remove(armor1FullPath);
            if (fileExists(armor1FullPath))
            {
                printf("\nFailed to delete %s", armor1FullPath);
                PAUSE();
            }

        }

        if (rename(armor2FullPath,armor1FullPath) == 0)
        {
            if (DEBUG)
            {
                printf("\nSuccessfully renamed:\n\"%s\" \nto \n\"%s\"\n", armor2FullPath, armor1FullPath);
            }
        }
        else
        {
             printf("\n\nFailed to rename:\n\"%s\" \nto \n\"%s\"\n", armor2FullPath, armor1FullPath);
             printf("--------------------------------------------------");
             errors++;
        }
    }
    printf("\n\n");
    applyPatch(partsToPatch1, partsToPatchCount, metaFileInfo);
}

int main()
{
    system("patcher_resources\\armors_to_swap.txt");
    MetaFileInfo *metaFileInfo = getMetaFileInfo(getLatestBackup());
    FileBlock* fileBlocks = fillFileBlocks(metaFileInfo);

    FILE* armor_to_swap_file = openFile("patcher_resources\\armors_to_swap.txt","rb");
    createPath(concatenate(bdoRootFolder(),"object\\texture\\"));
    copyFileTo("blank_dxt1.dds","common_decals_road_04_dec.dds","patcher_resources\\texture\\",concatenate(bdoRootFolder(),"object\\texture\\"),1);

    char armor1[128] = "";
    char armor2[128] = "";
    char c;
    int i = 0;
    int arrowFound = 0;
    int readingArmorNumber = 1;
    int result = 0;


    while (1)
    {
        result = fread(&c,1,1,armor_to_swap_file);
        // Skip Comments
        while (c == '#')
        {
            while(c != '\n')
            {
                fread(&c,1,1,armor_to_swap_file);
            }
            fread(&c,1,1,armor_to_swap_file);
        }

        if (readingArmorNumber == 1 && c != ' ' && c != '\n')
        {
            armor1[i] = c;
            i++;
        }
        else if (readingArmorNumber == 2 && c != ' ' && c != '\n')
        {
            armor2[i] = c;
            i++;
        }

        if (c == '\n' && arrowFound == 0)
        {
            //printf("\nSymbol \"->\" not found");
            i = 0;
        }
        if (c == '-')
        {
            fread(&c,1,1,armor_to_swap_file);
            if (c == '>')
            {
                armor1[i-1] = '\0';
                arrowFound = 1;
                readingArmorNumber = 2;

            }
            i = 0;
        }
        if ((c == '\n' && arrowFound == 1) || (result == 0 && arrowFound == 1))
        {
            armor2[i-1] = '\0';

            if (strcmpi(armor1,"ARMOR_TO_BE_REPLACED.PAC") != 0 && strcmpi(armor2,"ARMOR_TO_REPLACE_WITH.PAC") != 0)
            {
                printf("---------------------------------------------");
                printf("\narmor1: %s", armor1);
                printf("\narmor2: %s\n\n", armor2);

                if (!binarySearchFileBlock(armor1,fileBlocks, metaFileInfo->fileBlocksCount))
                {
                    printf("\nError: \"%s\" was not found in the reference file.\n\n", armor1);
                    errors++;
                    system("PAUSE");
                }
                else if (!binarySearchFileBlock(armor2,fileBlocks, metaFileInfo->fileBlocksCount))
                {
                    printf("\nError: \"%s\" was not found in the reference file.\n\n", armor2);
                    errors++;
                    system("PAUSE");
                }
                else
                {


                    swapArmors(armor1,armor2,fileBlocks,metaFileInfo);

                    if(errors != 0)
                    {
                        printf("\nSome errors have occurred...\nPress any key to run again with Debug Info ON for more detailed information.\n\n");
                        system("PAUSE");
                        DEBUG = 1;
                        system("cls");
                        swapArmors(armor1,armor2,fileBlocks,metaFileInfo);
                        DEBUG = 0;
                    }
                    else
                    {
                        success++;
                    }
                }


            }

            arrowFound = 0;
            readingArmorNumber = 1;
            i = 0;
            armor1[0] = '\0';
            armor2[0] = '\0';
            errors = 0;
            totalSwaps++;

        }
        if (result == 0)
        {
            break;
        }
    }
    if (totalSwaps == 0)
    {
        printf("\nNo swaps detected in the armors_to_swap.txt file.");
    }
    displayResults();
    printf("\n\n");
    system("PAUSE");
    return 0;
}



void displayResults()
{
    printColor("\n\n-------------- Overall Results --------------\n", YELLOW);
    printf("Armors Swapped: \033[32;1m%d\033[0m Not Swapped: \033[31;1m%d\033[0m",success, (int)(totalSwaps - success));

    if (totalSwaps == success)
    {
        printColor("\n\n\tAll armors have been swapped successfully!", GREEN);
    }
}

