#include "../include/patcher.h"

/* **********************************************************************
 *
 *                        RUN PATCHER
 *
 ************************************************************************/
void runPatcher(FileBlock* fileBlocks, MetaFileInfo* metaFileInfo,PatcherState* patcherState)
{
    FileBlock* filesToPatch = NULL;
    int filesToPatchCount = 0;

    // Checks if the files in the patcher_resources folder can be found in the File Blocks
    ///validatePatcherResourcesFiles(fileBlocks, metaFileInfo);

    // Allocates memory for the files to patch, worst case scenario, it will patch all the files so we allocate the same size as the fileBlocks
    filesToPatch = (FileBlock*) calloc(metaFileInfo->fileBlocksCount, sizeof(FileBlock));

    markFilesToPatch(filesToPatch, &filesToPatchCount, fileBlocks, metaFileInfo, patcherState);

     // QUICK SORT
    qsort(filesToPatch,filesToPatchCount,sizeof(FileBlock),compare_middle_digits);

    printf("\n------------- All set -------------\n");
    printf("\n%d file(s) will be patched.\n\n", filesToPatchCount);
    printf("This Option will patch your game according to the \n");
    printf("configuration set in the \"Customize\" and other menu options.\n\n");
    printf("Do you want to proceed or you want to go back and change your settings first?\n");
    printf("ENTER = Proceed    ESC = Go Back\n");
    char c = '\0';

    while(c != ENTER &&  c != 10 && c != ESC)
    {
        c = getch();
    }
    if (c == ESC)
    {
        return;
    }

    copyPatcherResourcesFiles(filesToPatch, filesToPatchCount, patcherState);

    if (bodyPartsSizeChanged(patcherState))
    {
        printf("\nModifying XML files...\n");
        patchBreastSizePatcherFiles(filesToPatch, filesToPatchCount, patcherState);
        printf("Done.\n");
    }

    printf("\n\nCopying nude textures\n");
    copyNudeTextures(patcherState);
    printf("Done.\n");

    undoLastChanges(0);
    patchMetaFile( filesToPatch, filesToPatchCount, INSTALL, metaFileInfo,0);
    filesPatched = countPatchedFiles(fileBlocks,metaFileInfo);

    free(filesToPatch);

    printf("\n\nFinished!\n");
    system("PAUSE");
}


void markFilesToPatch(FileBlock* filesToPatch, int* filesToPatchCount, FileBlock* fileBlocks, MetaFileInfo* metaFileInfo, PatcherState* patcherState)
{
    printf("\nMarking files to patch...\n");
    int i = 0;
    int j = 0;
    clearFileBlocks(fileBlocks,metaFileInfo);

    for (i = 0; i < metaFileInfo->fileBlocksCount; i++)
    {
        // Clears the need patch variable in case the user runs the program multiple times
        filesToPatch[i].needPatch = 0;

        // This is the texture of the pac file that every removed armor will load, it's now a blank texture so you can't see it
        if (strcmpi(fileBlocks[i].fileName,"common_decals_road_04_dec.dds") == 0)
        {
            addToFilesToPatch(&fileBlocks[i],&filesToPatch[(*filesToPatchCount)],filesToPatchCount);
        }

        // Breast Size Patcher
        if (contains(fileBlocks[i].fileName,"customizationboneparamdesc"))
        {
            addToFilesToPatch(&fileBlocks[i],&filesToPatch[(*filesToPatchCount)],filesToPatchCount);
        }

        // Transparent Kibelius
         if (patcherState->transparentKibelius != 0)
        {
            if ((strcmpi(fileBlocks[i].fileName,"pew_00_ub_0034_hair.dds") == 0) || (strcmpi(fileBlocks[i].fileName,"pew_00_ub_0034_01_hair.dds") == 0) || (strcmpi(fileBlocks[i].fileName,"pew_00_ub_0034_dm_hair.dds") == 0) || (strcmpi(fileBlocks[i].fileName,"pew_00_ub_0034_01_dm_hair.dds") == 0))
            {
                addToFilesToPatch(&fileBlocks[i],&filesToPatch[(*filesToPatchCount)],filesToPatchCount);
            }
        }

        // WEAPON REMOVAL
        if (contains(fileBlocks[i].folderName,"weapon") && contains(fileBlocks[i].folderName,"1_pc") && fileHasExt(fileBlocks[i].fileName,"pac") && !contains(fileBlocks[i].folderName,"0_com") )
        {
            weaponOptionRequirePatch(&fileBlocks[i], patcherState);
        }

        //  Valkyrie Venslar Feather removal
        if (patcherState->option[ARMOR] == REMOVE_ALL_ARMORS && strcmpi(fileBlocks[i].fileName,"pvw_00_feather_0001_hair.dds") == 0)
        {
            fileBlocks[i].needPatch = 1;
        }

        // ALL OTHER FILES
        if (((patcherState->option[GENDER] == FEMALE_ONLY) && (isAFemaleFile(fileBlocks[i].fileName)))
        ||  ((patcherState->option[GENDER] == MALE_ONLY  ) && (isAMaleFile(fileBlocks[i].fileName)))
        ||   (patcherState->option[GENDER] == BOTH_GENDERS))
        {
                // If it's a .pac file
                if (fileHasExt(fileBlocks[i].fileName,"pac"))
                {
                    if (contains(fileBlocks[i].folderName,"9_upperbody")
                    || contains(fileBlocks[i].folderName,"14_sho")
                    || contains(fileBlocks[i].folderName,"19_cloak"))
                    {
                        if (patcherState->option[ARMOR] == REMOVE_ALL_ARMORS)
                        {
                            fileBlocks[i].needPatch = 1;
                            fileBlocks[i].changeID = 1;
                        }
                        else if  (patcherState->option[ARMOR] == REMOVE_STARTER_ARMORS)
                        {
                            for (j = 0; j < TOTAL_CLASSES; j++)
                            {
                                if (strcmpi(fileBlocks[i].fileName,starters[ARMOR][j]) == 0
                                || strcmpi(fileBlocks[i].fileName,starters[SHOULDER][j]) == 0)
                                {
                                    fileBlocks[i].needPatch = 1;
                                    fileBlocks[i].changeID = 2;
                                }
                            }
                        }
                        else if  (patcherState->option[ARMOR] == REMOVE_REGULAR_ARMORS)
                        {
                           if (isRegular(fileBlocks[i].fileName))
                           {
                               fileBlocks[i].needPatch = 1;
                               fileBlocks[i].changeID = 3;
                           }
                        }

                        // Prevents Kibelius armor  from being patch if it's set to transparent
                        if (patcherState->transparentKibelius != 0)
                        {
                            if (contains(fileBlocks[i].fileName,"_00_ub_0034"))
                            {
                               fileBlocks[i].needPatch = 0;
                            }
                        }


                        if (contains(fileBlocks[i].fileName,"_00_sho_0034"))
                        {
                            if (patcherState->kibeliusWings == KEEP_WINGS)
                            {
                                fileBlocks[i].needPatch = 0;
                            }
                            else
                            {
                                fileBlocks[i].needPatch = 1;
                            }
                        }
                    }

                    else if (contains(fileBlocks[i].folderName,"event_costume"))
                    {
                        if  (patcherState->option[LIFE_SKILL_ARMORS] == REMOVE_ALL_LIFE_SKILL)
                        {
                            fileBlocks[i].needPatch = 1;
                            fileBlocks[i].changeID = 4;
                        }
                    }

                   /* if (contains(fileBlocks[i].folderName,"19_cloak"))
                    {
                         if  (patcherState->option[LIFE_SKILL_ARMORS] == REMOVE_ALL_LIFE_SKILL)
                         {
                             char* twoDigits = getTwoDigits(fileBlocks[i].fileName);
                             char* midleDigits = getMiddleDigits(fileBlocks[i].fileName);
                             char* fourDigits = getFourDigits(fileBlocks[i].fileName);

                            if ((strcmpi(midleDigits,"TreeDummer") == 0)
                            || (strcmpi(twoDigits,"10") == 0 && strcmpi(fourDigits,"0004") == 0))
                            {
                                fileBlocks[i].needPatch = 1;
                                fileBlocks[i].changeID = 7;
                            }
                         }
                    }*/

                    else if (contains(fileBlocks[i].folderName,"38_underwear"))
                    {

                        if (patcherState->option[UNDERWEAR] == REMOVE_ALL_UNDERWEAR)
                        {
                            fileBlocks[i].needPatch = 1;
                            fileBlocks[i].changeID = 8;
                        }
                        else if  (patcherState->option[UNDERWEAR] == REMOVE_STARTER_UNDERWEAR)
                        {
                            if (contains(fileBlocks[i].fileName,"_00_uw_0001"))
                            {
                                fileBlocks[i].needPatch = 1;
                                fileBlocks[i].changeID = 43;
                            }
                        }

                        if (patcherState->option[STOCKINGS] == KEEP_ALL_STOCKINGS || patcherState->option[STOCKINGS] == REMOVE_STARTER_UNDERWEAR)
                        {
                            if (isAFemaleFile(fileBlocks[i].fileName))
                            {
                                fileBlocks[i].needPatch = 0;
                            }


                           /* char* fileEnding = substr(fileBlocks[i].fileName,3,strlen(fileBlocks[i].fileName));

                            if (patcherState->option[UNDERWEAR] == REMOVE_ALL_UNDERWEAR)
                            {
                                if ((strcmpi(fileEnding,"_00_uw_0035.pac") == 0) || (strcmpi(fileEnding,"_00_uw_0004.pac") == 0))
                                {
                                    fileBlocks[i].needPatch = 1;
                                    fileBlocks[i].changeID = 44;
                                }
                            }



                            if (patcherState->option[UNDERWEAR] != KEEP_ALL_UNDERWEAR)
                            {
                                if (strcmpi(fileBlocks[i].fileName,"pkww_00_uw_0001.pac") == 0)
                                {
                                     fileBlocks[i].needPatch = 1;
                                     fileBlocks[i].changeID = 45;
                                }
                            }
                            // REPLACES UB WITH UNDERWEAR
                            if (patcherState->option[UNDERWEAR] == KEEP_ALL_UNDERWEAR && patcherState->option[ARMOR] != KEEP_ALL_ARMORS)
                            {
                                fileBlocks[i].needPatch = 1;
                                fileBlocks[i].changeID = 46;
                            }*/
                        }
                    }

                    if (contains(fileBlocks[i].folderName,"10_lowerbody"))
                    {
                        if (patcherState->option[ARMOR] == REMOVE_ALL_ARMORS)
                        {
                            fileBlocks[i].needPatch = 1;
                            fileBlocks[i].changeID = 47;
                        }
                        if (patcherState->option[ARMOR] == REMOVE_STARTER_ARMORS)
                        {
                            for (j = 0; j < TOTAL_CLASSES; j++)
                            {
                                if (strcmpi(fileBlocks[i].fileName,starters[LOWERBODY][j]) == 0)
                                {
                                    fileBlocks[i].needPatch = 1;
                                    fileBlocks[i].changeID = 9;
                                }
                            }
                        }
                        if  (patcherState->option[ARMOR] == REMOVE_REGULAR_ARMORS)
                        {
                           if (isRegular(fileBlocks[i].fileName))
                           {
                               fileBlocks[i].needPatch = 1;
                               fileBlocks[i].changeID = 10;
                           }
                        }
                        if (patcherState->option[LOWERBODY] == REMOVE_PANTIES_AND_PANTS)
                        {
                            for (j = 0; j < TOTAL_PANTIES; j++)
                            {
                                if (strcmpi(fileBlocks[i].fileName,panties[j]) == 0)
                                {
                                    fileBlocks[i].needPatch = 1;
                                    fileBlocks[i].changeID = 11;
                                }
                            }
                            for (j = 0; j < TOTAL_PANTS; j++)
                            {
                                if (strcmpi(fileBlocks[i].fileName,pants[j]) == 0)
                                {
                                   fileBlocks[i].needPatch = 1;
                                   fileBlocks[i].changeID = 12;
                                }
                            }
                        }
                        if  (patcherState->option[LOWERBODY] == REMOVE_STARTER_PANTIES)
                        {
                            for (j = 0; j < TOTAL_CLASSES; j++)
                            {
                                if (strcmpi(fileBlocks[i].fileName,starters[LOWERBODY][j]) == 0)
                                {
                                    fileBlocks[i].needPatch = 1;
                                    fileBlocks[i].changeID = 13;
                                }
                            }
                        }
                        if  (patcherState->option[LOWERBODY] == REMOVE_PANTIES_ONLY)
                        {
                            for (j = 0; j < TOTAL_PANTIES; j++)
                            {
                                if (strcmpi(fileBlocks[i].fileName,panties[j]) == 0)
                                {
                                    fileBlocks[i].needPatch = 1;
                                    fileBlocks[i].changeID = 14;
                                }
                            }
                        }
                        if  (patcherState->option[LOWERBODY] == REMOVE_PANTS_ONLY)
                        {
                             for (j = 0; j < TOTAL_PANTS; j++)
                            {
                                if (strcmpi(fileBlocks[i].fileName,pants[j]) == 0)
                                {
                                   fileBlocks[i].needPatch = 1;
                                   fileBlocks[i].changeID = 15;
                                }
                            }
                        }
                        if  (patcherState->option[LOWERBODY] == REMOVE_ALL_PANTIES_AND_PANTS)
                        {
                            fileBlocks[i].needPatch = 1;
                            fileBlocks[i].changeID = 16;
                        }
                    }

                    if (contains(fileBlocks[i].folderName,"11_hand"))
                    {
                        if (patcherState->option[GLOVES] == REMOVE_ALL_GLOVES)
                        {
                            fileBlocks[i].needPatch = 1;
                            fileBlocks[i].changeID = 17;
                        }
                        else if  (patcherState->option[GLOVES] == REMOVE_STARTER_GLOVES)
                        {
                            for (j = 0; j < TOTAL_CLASSES; j++)
                            {
                                if (strcmpi(fileBlocks[i].fileName,starters[GLOVES][j]) == 0)
                                {
                                    fileBlocks[i].needPatch = 1;
                                    fileBlocks[i].changeID = 18;
                                }
                            }
                        }
                        else if  (patcherState->option[GLOVES] == REMOVE_REGULAR_GLOVES)
                        {
                           if (isRegular(fileBlocks[i].fileName))
                           {
                               fileBlocks[i].needPatch = 1;
                               fileBlocks[i].changeID = 19;
                           }
                        }

                        if (contains(fileBlocks[i].fileName,"_10_hand_0003.pac"))
                        {
                            fileBlocks[i].needPatch = 0;
                            fileBlocks[i].changeID = 20;
                        }
                    }


                    if (contains(fileBlocks[i].folderName,"12_foot"))
                    {
                        if (patcherState->option[BOOTS] == REMOVE_ALL_BOOTS)
                        {
                             fileBlocks[i].needPatch = 1;
                             fileBlocks[i].changeID = 21;
                        }
                        else if  (patcherState->option[BOOTS] == REMOVE_STARTER_BOOTS)
                        {
                            for (j = 0; j < TOTAL_CLASSES; j++)
                            {
                                if (strcmpi(fileBlocks[i].fileName,starters[BOOTS][j]) == 0)
                                {
                                    fileBlocks[i].needPatch = 1;
                                    fileBlocks[i].changeID = 22;
                                }
                            }
                        }
                        else if  (patcherState->option[BOOTS] == REMOVE_REGULAR_BOOTS)
                        {
                           if (isRegular(fileBlocks[i].fileName))
                           {
                               fileBlocks[i].needPatch = 1;
                               fileBlocks[i].changeID = 23;
                           }
                        }

                        if (contains(fileBlocks[i].fileName,"_10_foot_0003.pac"))
                        {
                            fileBlocks[i].needPatch = 0;
                            fileBlocks[i].changeID = 24;
                        }
                    }

                    if (contains(fileBlocks[i].folderName,"13_hel"))
                    {
                        if (patcherState->option[HELMETS] == REMOVE_ALL_HELMETS)
                        {
                            int end = indexOf('.',fileBlocks[i].fileName,0);
                            int start = end - strlen("haircut");
                            char* sub = substr(fileBlocks[i].fileName,start,end);

                            if ((strcmpi(sub,"haircut") != 0))
                            {
                                if(isAFemaleFile(fileBlocks[i].fileName))
                                {
                                    if (!contains(fileBlocks[i].fileName,"_00_hel_0052")
                                    && !contains(fileBlocks[i].fileName,"_10_hel_0008")
                                    && !contains(fileBlocks[i].fileName,"_00_hel_0072")
                                    && !contains(fileBlocks[i].fileName,"_00_hel_0083")
                                    && !contains(fileBlocks[i].fileName,"_00_hel_0053")
                                    && !contains(fileBlocks[i].fileName,"_00_hel_0037"))
                                    {
                                            fileBlocks[i].needPatch = 1;
                                            fileBlocks[i].changeID = 25;
                                    }
                                }
                            }
                        }
                    }
                    if (contains(fileBlocks[i].folderName,"Hair"))
                    {
                        if (patcherState->option[HELMETS] == REMOVE_ALL_HELMETS)
                        {
                            int start = indexOf('_',fileBlocks[i].fileName,1) + 1;
                            int end = indexOf('_',fileBlocks[i].fileName,3);
                            char* sub = substr(fileBlocks[i].fileName,start,end);

                            if (strcmpi(sub,"Hair_Hel") == 0)
                            {
                                fileBlocks[i].needPatch = 0;
                                fileBlocks[i].changeID = 26;
                            }
                        }
                    }

                    if (isATreeDummerFile(fileBlocks[i].fileName))
                    {
                        fileBlocks[i].needPatch = 0;
                    }
                }

                /****************************************************************************
                 *                            TEXTURE PATCHING                              *
                 ****************************************************************************/
                if (fileHasExt(fileBlocks[i].fileName,"dds"))
                {
                    if (/**!*/!(patcherState->option[UNDERWEAR] == REMOVE_ALL_UNDERWEAR && patcherState->option[STOCKINGS] == REMOVE_ALL_STOCKINGS) || contains(fileBlocks[i].fileName,"_ao.dds"))
                    {
                        if (contains(fileBlocks[i].folderName,"texture"))
                        {
                            if (contains(fileBlocks[i].fileName,"_uw_") || isStockingTexture(fileBlocks[i].fileName) || contains(fileBlocks[i].fileName,"_99_ub_"))
                            {
                                if(!contains(fileBlocks[i].fileName,"_n.dds")
                                   && !contains(fileBlocks[i].fileName,"_sp.dds")
                                   && !contains(fileBlocks[i].fileName,"_m.dds")
                                   && !contains(fileBlocks[i].fileName,"_st.dds")
                                   && !contains(fileBlocks[i].fileName,"_fur"))
                                {

                                    // UPPER BODY UNDERWEAR PATCH
                                    if (patcherState->option[UNDERWEAR] == REMOVE_ALL_UNDERWEAR)
                                    {
                                        if (isStockingTexture(fileBlocks[i].fileName) != 1)
                                        {
                                            fileBlocks[i].needPatch = 1;
                                            fileBlocks[i].changeID = 36;
                                        }

                                    }
                                     if (patcherState->option[UNDERWEAR] == REMOVE_STARTER_UNDERWEAR)
                                    {
                                        for (j = 0; j < TOTAL_STARTER_UNDERWEAR_TEXTURES; j++)
                                        {
                                            if (strcmpi(fileBlocks[i].fileName,starterUnderwearTextures[j]) == 0)
                                            {
                                                fileBlocks[i].needPatch = 1;
                                                fileBlocks[i].changeID = 39;
                                            }
                                        }
                                    }

                                    //STOCKINGS PATCH
                                    if (patcherState->option[STOCKINGS] == REMOVE_ALL_STOCKINGS)
                                    {
                                        if (isStockingTexture(fileBlocks[i].fileName))
                                        {
                                            // Patch the upper body part
                                            fileBlocks[i].needPatch = 1;
                                            fileBlocks[i].changeID = 35;
                                        }
                                    }
                                     if (patcherState->option[STOCKINGS] == KEEP_ALL_STOCKINGS)
                                    {
                                        if (isStockingTexture(fileBlocks[i].fileName))
                                        {
                                            fileBlocks[i].needPatch = 0;
                                        }
                                    }
                                     if (patcherState->option[STOCKINGS] == REMOVE_STARTER_STOCKINGS)
                                    {
                                        for(j = 0; j < TOTAL_STARTER_STOCKINGS; j++)
                                        {
                                            if (strcmpi(fileBlocks[i].fileName,starterStockings[j]) == 0)
                                            {
                                                fileBlocks[i].needPatch = 1;
                                                fileBlocks[i].changeID = 41;
                                            }
                                        }
                                    }

                                }
                            }
                        }
                    }

                }
        }
              // Always patch AO files
        if (contains(fileBlocks[i].fileName,"_ao.dds"))
        {
            if (contains(fileBlocks[i].fileName,"_uw_") || contains(fileBlocks[i].fileName,"_99_ub_"))
            {
                fileBlocks[i].needPatch = 1;
            }
        }
        // Forces the nude textures to patch no matter what
        if (isANudeTexture(fileBlocks[i].fileName))
        {
            fileBlocks[i].needPatch = 1;
            fileBlocks[i].changeID = 42;
        }
        // Adds the file block to the files to patch array, if it reaches this point with needPatch == 1
        if (fileBlocks[i].needPatch == 1)
        {
             addToFilesToPatch(&fileBlocks[i],&filesToPatch[(*filesToPatchCount)],filesToPatchCount);
        }
    }
    printf("Done.\n");
}

void patchMetaFile(FileBlock* filesToPatch, int filesToPatchCount, int operation, MetaFileInfo* metaFileInfo,int echo)
{
    int i = 0;
    int filesPatched = 0;
    int filesFailedToPatch = 0;
    long random_folder_num = RANDOM_FOLDER_NUM;
    long random_file_num = RANDOM_FILE_NUM;
    int firstFail = 0;
    long hash_read_from_file = 0;

   /* printf("\n\nWe are about to make changes to your pad00000.meta file.");
    printf("\nPlease make sure you have made a manual backup before proceeding...\n\n");

    system("PAUSE");*/

    printf("\n\nPatching Meta File...\n");

    FILE* metaFile = openFile("pad00000.meta","rb+");

    FILE* modFile = NULL;

    if (operation == INSTALL)
    {
       modFile = fopen(concatenate("patcher_resources\\",LATEST_MODIFICATIONS_FILE_NAME),"wb+");
    }
    /*else // If operation == RESTORE
    {
        modFile = fopen(concatenate("patcher_resources\\",LATEST_MODIFICATIONS_FILE_NAME),"ab+"); // Append to the end of the file
    }*/

    for (i = 0; i < filesToPatchCount; i++)
    {
        if(filesToPatch[i].needPatch == 1 && operation != RESTORE_A_FILE)
        {
            fseek(metaFile,filesToPatch[i].metaOffset, SEEK_SET); // Goes to the position the File Block starts

            fread(&hash_read_from_file,sizeof(long),1,metaFile);

            if (hash_read_from_file == filesToPatch[i].hash)
            {
                if (modFile != NULL)
                {
                    // Backup the original values to the  latest modification file
                    fwrite(&filesToPatch[i].metaOffset,sizeof(long),1,modFile);
                    fwrite(&filesToPatch[i].hash,sizeof(long),1,modFile);
                    fwrite(&filesToPatch[i].folderNum,sizeof(long),1,modFile);
                    fwrite(&filesToPatch[i].fileNum,sizeof(long),1,modFile);

                    int fileNameLen = strlen(filesToPatch[i].fileName);
                    fwrite(&fileNameLen,sizeof(int),1,modFile);

                    fwrite(filesToPatch[i].fileName,fileNameLen,sizeof(char),modFile);
                }

                if (echo)
                {

                    if (firstFail == 0)
                    {
                        printf("          HASH  FOLDER_NUM FILE_NUM   OFFSET   | FILE_NAME\n");
                        firstFail = 1;
                    }
                    printColor("Patched ", GREEN);
                    printf("%s %s %s  (%.8ld) | %s\n", endian_convert((int)filesToPatch[i].hash),endian_convert((int)filesToPatch[i].folderNum),endian_convert((int)filesToPatch[i].fileNum), filesToPatch[i].metaOffset, filesToPatch[i].fileName);
               }

                // Goes to the position the Folder Num
                fseek(metaFile,filesToPatch[i].metaOffset + sizeof(long), SEEK_SET);

                /// The most important part, breaks the meta file index by changing the 2 next ints after the hash with any random value that doesn't have a folderNum && fileNum match
                fwrite(&random_folder_num, sizeof(long),1,metaFile);
                fwrite(&random_file_num, sizeof(long),1,metaFile);
                filesPatched++;
            }
            else /*** FAIL: IF THE HASH WAS NOT WHERE IT WAS SUPOSED TO BE */
            {
               int found = 0;
               fseek(metaFile,metaFileInfo->fileBlocksStart,SEEK_SET);
               while(ftell(metaFile) < metaFileInfo->fileBlocksEnd)
               {
                    fread(&hash_read_from_file,sizeof(long),1,metaFile);
                    if (hash_read_from_file == filesToPatch[i].hash)
                    {
                        /** FAIL PART */
                        if (modFile != NULL)
                        {
                            // Backup the original values to the  latest modification file
                            fwrite(&filesToPatch[i].metaOffset,sizeof(long),1,modFile);
                            fwrite(&filesToPatch[i].hash,sizeof(long),1,modFile);
                            fwrite(&filesToPatch[i].folderNum,sizeof(long),1,modFile);
                            fwrite(&filesToPatch[i].fileNum,sizeof(long),1,modFile);

                            int fileNameLen = strlen(filesToPatch[i].fileName);
                            fwrite(&fileNameLen,sizeof(int),1,modFile);

                            fwrite(filesToPatch[i].fileName,fileNameLen,sizeof(char),modFile);
                        }

                        // Goes to the position the Folder Num
                        fseek(metaFile,filesToPatch[i].metaOffset + sizeof(long), SEEK_SET);


                        /// Patches the meta file
                        fwrite(&random_folder_num, sizeof(long),1,metaFile);
                        fwrite(&random_file_num, sizeof(long),1,metaFile);

                       if (firstFail == 0)
                       {
                            printf("          HASH  FOLDER_NUM FILE_NUM   OFFSET   | FILE_NAME\n");
                            firstFail = 1;
                       }

                        printColor("Patched ", YELLOW);
                        printf("%s %s %s  (%.8ld) | %s\n", endian_convert((int)filesToPatch[i].hash),endian_convert((int)filesToPatch[i].folderNum),endian_convert((int)filesToPatch[i].fileNum), filesToPatch[i].metaOffset, filesToPatch[i].fileName);
                        found = 1;

                        filesPatched++;
                        break;
                    }
                    else
                    {
                        // Skips from File Blocks to file blocks
                        //fseek(metaFile,ONE_REGISTRY - sizeof(long),SEEK_CUR);
                    }
               }
               if (!found)
               {
                   if (firstFail == 0)
                   {
                        printf("          HASH  FOLDER_NUM FILE_NUM   OFFSET   | FILE_NAME\n");
                        firstFail = 1;
                   }
                   printColor(" Failed ", RED);
                   printf("%s %s %s  (%.8ld) | %s\n",endian_convert((int)filesToPatch[i].hash),endian_convert((int)filesToPatch[i].folderNum),endian_convert((int)filesToPatch[i].fileNum), filesToPatch[i].metaOffset, filesToPatch[i].fileName);
                   filesToPatch[i].changeID = -2;
                   filesFailedToPatch++;
               }
            }
        }
        if(filesToPatch[i].needPatch == 1 && operation == RESTORE_A_FILE)
        {
            fseek(metaFile,filesToPatch[i].metaOffset + sizeof(long), SEEK_SET); // Goes to the position right after the Hash, which has the folder num value

            // Restore
            fwrite(&filesToPatch[i].folderNum, sizeof(long),1,metaFile);
            fwrite(&filesToPatch[i].fileNum, sizeof(long),1,metaFile);


            if (echo)
            {

                if (firstFail == 0)
                {
                    printf("          HASH  FOLDER_NUM FILE_NUM   OFFSET   | FILE_NAME\n");
                    firstFail = 1;
                }
                printColor("Restored ", YELLOW);
                printf("%s %s %s  (%.8ld) | %s\n", endian_convert((int)filesToPatch[i].hash),endian_convert((int)filesToPatch[i].folderNum),endian_convert((int)filesToPatch[i].fileNum), filesToPatch[i].metaOffset, filesToPatch[i].fileName);
           }
           else
           {
                printColor("Restored ", YELLOW);
           }
        }
    }

    if (filesFailedToPatch > 0)
    {
        printColor("\nThese files failed to patch:\n",RED);
        for(i = 0; i < filesToPatchCount; i++)
        {
            if (filesToPatch[i].changeID == -2)
            {
                printf("%s\n", filesToPatch[i].fileName);
            }
        }
        printColor("\nTotal failed:",RED);printf(" %d", filesFailedToPatch);
    }
    else
    {
        printf("Success.\n");
    }


    fclose(metaFile);
    fclose(modFile);
}

void copyPatcherResourcesFiles(FileBlock* filesToPatch, int filesToPatchCount, PatcherState* patcherState)
{
    printf("\nCopying Files...\n");
    int i = 0;
    int extractedBreastSizePatcherFiles = 0;
    char* root = bdoRootFolder();

    char* old_class_prefix = (char*) malloc(1);
    old_class_prefix[0] = '\0';
    char* newclass_prefix = NULL;

    for (i = 0; i < filesToPatchCount; i++)
    {
        printFileBeingCopied(&filesToPatch[i], filesToPatchCount, &newclass_prefix, &old_class_prefix, i);

        // PAC FILE COPY
        if (fileHasExt(filesToPatch[i].fileName,"pac"))
        {
            // Replace every pac file marked to patch, with a ground model, which we removed the texture, so it's transparent
            copyFileTo("t0072_pumpkin_ground_0001.pac",filesToPatch[i].fileName,"patcher_resources/models/",concatenate(root,filesToPatch[i].folderName),0);
        }
        // DDS FILE COPY
        else if (fileHasExt(filesToPatch[i].fileName,"dds"))
        {
            // If it's the Transparent Kibelius textures
            if ((strcmpi(filesToPatch[i].fileName,"pew_00_ub_0034_hair.dds") == 0)
                || (strcmpi(filesToPatch[i].fileName,"pew_00_ub_0034_01_hair.dds") == 0)
                || (strcmpi(filesToPatch[i].fileName,"pew_00_ub_0034_dm_hair.dds") == 0)
                || (strcmpi(filesToPatch[i].fileName,"pew_00_ub_0034_01_dm_hair.dds") == 0))
            {
                if (patcherState->transparentKibelius != 0)
                {
                    if (patcherState->transparentKibelius == TRANSPARENT)
                    {
                         copyFileTo(filesToPatch[i].fileName,filesToPatch[i].fileName,"patcher_resources\\texture\\kibelius\\transparent\\",concatenate(root,filesToPatch[i].folderName),0);
                    }
                    else if (patcherState->transparentKibelius == MORE_TRANSPARENT)
                    {
                         copyFileTo(filesToPatch[i].fileName,filesToPatch[i].fileName,"patcher_resources\\texture\\kibelius\\more_transparent\\",concatenate(root,filesToPatch[i].folderName),0);
                    }
                }
            }
             else // Regular DDS copy
            {
                if (!contains(filesToPatch[i].fileName,"_ao.dds"))
                {
                    if (filesToPatch[i].size < 1000000)
                    {
                        copyFileTo("blank_dxt1.dds",filesToPatch[i].fileName,"patcher_resources/texture/",concatenate(root,filesToPatch[i].folderName),0);
                    }
                    else
                    {
                        copyFileTo("blank_dxt5.dds",filesToPatch[i].fileName,"patcher_resources/texture/",concatenate(root,filesToPatch[i].folderName),0);
                    }
                }
                else
                {
                    if (strcmpi(filesToPatch[i].fileName,"pdw_00_uw_0001_ao.dds") == 0)
                    {
                        copyFileTo("pdw_00_uw_0001_ao.dds",filesToPatch[i].fileName,"patcher_resources/texture/",concatenate(root,filesToPatch[i].folderName),0);
                    }
                    else if (filesToPatch[i].size >= 600000)
                    {
                        copyFileTo("blank_ao_683.dds",filesToPatch[i].fileName,"patcher_resources/texture/",concatenate(root,filesToPatch[i].folderName),0);
                    }
                    else if (filesToPatch[i].size >= 170000 && filesToPatch[i].size < 600000)
                    {
                        copyFileTo("blank_ao_171.dds",filesToPatch[i].fileName,"patcher_resources/texture/",concatenate(root,filesToPatch[i].folderName),0);
                    }
                     else if (filesToPatch[i].size >= 43000 && filesToPatch[i].size < 170000)
                    {
                        copyFileTo("blank_ao_43.dds",filesToPatch[i].fileName,"patcher_resources/texture/",concatenate(root,filesToPatch[i].folderName),0);
                    }
                    else
                    {
                        printColor("        _ao file size mismatch\n",YELLOW);
                        copyFileTo("blank_ao_171.dds",filesToPatch[i].fileName,"patcher_resources/texture/",concatenate(root,filesToPatch[i].folderName),0);
                    }
                }
            }
        }

        // Breast Size Patcher files
        if (contains(filesToPatch[i].fileName,"customizationboneparamdesc"))
        {
            if (extractedBreastSizePatcherFiles == 0)
            {
                extractBreastSizePatcherFiles(filesToPatch,filesToPatchCount,0);
                extractedBreastSizePatcherFiles = 1;
            }
            checkExtractedFile(&filesToPatch[i]);
        }
    }

    free(newclass_prefix);
    free(old_class_prefix);
    free(root);
}



void printFileBeingCopied(FileBlock* fileToPatch,long filesToPatchCount, char** newclass_prefix, char** old_class_prefix, int i)
{
    (*newclass_prefix) = getClassPrefix(fileToPatch->fileName);

    if (strcmp((*newclass_prefix),(*old_class_prefix)) != 0)
    {
        // OUTPUT
        printf("\n********************************************");

        if (fileHasExt(fileToPatch->fileName,"dds"))
        {
             printf("\n%s %s Textures", getClassAsString(fileToPatch), getArmorPartAsString(fileToPatch));
        }
        else
        {
            printf("\n%s %s", getClassAsString(fileToPatch), getArmorPartAsString(fileToPatch));
        }

        free((*old_class_prefix));
        printf("\n********************************************\n");
        (*old_class_prefix) = (char*) malloc(strlen((*newclass_prefix)) + 1);
        strcpy((*old_class_prefix),(*newclass_prefix));
    }

    printf("(%d/%ld) - %s",i+1,filesToPatchCount,fileToPatch->fileName);
    //printf(" (%d)",  fileToPatch->changeID);
}

int isRestricted(int CLASS)
{
    if (classRestrictions[CLASS] == 1)
    {
        return 1;
    }
    return 0;
}




void addToFilesToPatch(FileBlock* fileBlockFound, FileBlock* filesToPatch, int* filesToPatchCount)
{
    filesToPatch->metaOffset = fileBlockFound->metaOffset;
    filesToPatch->hash = fileBlockFound->hash;
    filesToPatch->folderNum = fileBlockFound->folderNum;
    filesToPatch->fileNum = fileBlockFound->fileNum;
    filesToPatch->pazNum = fileBlockFound->pazNum;
    filesToPatch->fileOffset = fileBlockFound->fileOffset;
    filesToPatch->zsize = fileBlockFound->zsize;
    filesToPatch->size = fileBlockFound->size;
    filesToPatch->fileName = fileBlockFound->fileName;
    filesToPatch->folderName = fileBlockFound->folderName;
    filesToPatch->changeID = fileBlockFound->changeID;
    filesToPatch->needPatch = 1;
    (*filesToPatchCount)++;

}


int isRegular(char* fileName)
{
    char *exceptionsKeep[] =
    {
        "PHW_02_UB_0006.PAC","PHW_02_UB_0006_DM.PAC","PHW_02_LB_0006.PAC","PHW_02_SHO_0006.PAC","PHW_02_HAND_0006.PAC","PHW_02_FOOT_0006.PAC",
        "PHW_02_UB_0005.PAC","PHW_02_UB_0005_DM.PAC", "PHW_02_LB_0005.PAC", "PHW_02_SHO_0005.PAC","PHW_02_HAND_0005.PAC","PHW_02_FOOT_0005.PAC",
        "PEW_02_UB_0006.PAC","PEW_02_UB_0006_DM.PAC", "PEW_02_LB_0006.PAC", "PEW_02_SHO_0006.PAC", "PEW_02_HAND_0006.PAC","PEW_02_FOOT_0006.PAC"
    };

    char *exceptionsExclude[] =
    {
        "PVW_00_UB_0000.PAC","PVW_00_UB_0000_DM.PAC","PVW_00_LB_0000.PAC","PVW_00_SHO_0000.PAC","PVW_00_HAND_0000.PAC","PVW_00_FOOT_0000.PAC",
        "PVW_00_UB_0001.PAC", "PVW_00_UB_0001_DM.PAC","PVW_00_LB_0001.PAC","PVW_00_SHO_0001.PAC", "PVW_00_HAND_0001.PAC", "PVW_00_FOOT_0001.PAC",
        "PKWW_00_UB_0012.PAC","PKWW_00_UB_0012_DM.PAC", "PKWW_00_LB_0012.PAC","PKWW_00_SHO_0012.PAC","PKWW_00_HAND_0012.PAC","PKWW_00_FOOT_0012.PAC",
        "PKM_00_UB_0012.PAC", "PKM_00_UB_0012_DM.PAC", "PKM_00_LB_0012.PAC","PKM_00_SHO_0012.PAC","PKM_00_HAND_0012.PAC","PKM_00_FOOT_0012.PAC"
    };

    int i = 0;
    for (i = 0; i < 18; i++)
    {
        if ((strcmpi(fileName,exceptionsKeep[i]) == 0))
        {
            return 0;
        }
    }

    for (i = 0; i < 24; i++)
    {
        if ((strcmpi(fileName,exceptionsExclude[i]) == 0))
        {
            return 1;
        }
    }

    int start = indexOf('_',fileName,0) + 1;
    int end = indexOf('_',fileName,1);

    char* twoDigits = substr(fileName,start,end);


    if (((strcmpi(twoDigits,"01") == 0) || (strcmpi(twoDigits,"02") == 0 || strcmpi(twoDigits,"03") == 0)) && strlen(fileName) >= 18)
    {
        return 1;
    }
    return 0;
}

void copyNudeTextures(PatcherState* patcherState)
{
    int i = 0;
    char* textureDestination = concatenate(bdoRootFolder(),"character/texture/");

    for(i = 0; i < 5; i++)
    {
        createPath(textureDestination);

        systemCopy(nudeTextures[i],pubicHairFolder[patcherState->pubicHair[i]],textureDestination);

        char* srcPath = concatenate(getCurrentPath(),pubicHairFolder[patcherState->pubicHair[i]]);
        char* destPath = concatenate(textureDestination,nudeTextures[i]);
        if (!fileExists(destPath))
        {
            printColor("\nFailed to copy texture", RED);printf(" %s\n", nudeTextures[i]);
            printf("from: \"%s\"\n", srcPath);
            printf("to  : \"%s\"\n\n", textureDestination);
            printf("Make sure you have placed the texture in the correct location, or manually copy the texture to the destination.\n");
            printf("Warning: If you don't do anything, you will have floating heads with no body.");
            PAUSE();

        }
        free(srcPath);
        free(destPath);
    }
}

int isANudeTexture(char* fileName)
{
    int i = 0;
    for (i = 0; i < 5; i++)
    {
        if (strcmpi(fileName,nudeTextures[i]) == 0)
        {
            return 1;
        }
    }
    return 0;
}

// Checks if the files in the patcher_resources folder can be found in the File Blocks
void validatePatcherResourcesFiles(FileBlock* fileBlocks, MetaFileInfo* metaFileInfo)
{
    long modelFilesCount = 0;
    FileBlock* modelFiles = getAllFiles("patcher_resources\\models\\","*", &modelFilesCount);
    long textureFilesCount = 0;
    FileBlock* textureFiles = getAllFiles("patcher_resources\\texture\\","*", &textureFilesCount);

    int i = 0;
    int j = 0;
    for (i = 0; i < metaFileInfo->fileBlocksCount; i++)
    {
        for (j = 0; j < modelFilesCount; j++)
        {
            if (strcmpi(fileBlocks[i].fileName,modelFiles[j].fileName) == 0)
            {
                modelFiles[j].needPatch = 1;
            }
        }
        for (j = 0; j < textureFilesCount; j++)
        {
            if (strcmpi(fileBlocks[i].fileName,textureFiles[j].fileName) == 0)
            {
                textureFiles[j].needPatch = 1;
            }
        }
    }

    for (j = 0; j < modelFilesCount; j++)
    {
        if (modelFiles[j].needPatch != 1)
        {
            printf("\nNot found: %s", modelFiles[j].fileName);
        }
    }
    for (j = 0; j < textureFilesCount; j++)
    {
        if (textureFiles[j].needPatch != 1)
        {
            if (strcmpi(textureFiles[j].fileName,"blank_dxt1.dds") != 0
            && strcmpi(textureFiles[j].fileName,"blank_dxt5.dds") != 0
            && strcmpi(textureFiles[j].fileName,"blank_ao_683.dds") != 0
            && strcmpi(textureFiles[j].fileName,"blank_ao_171.dds") != 0
            && strcmpi(textureFiles[j].fileName,"blank_ao_43.dds") != 0
                && strcmpi(textureFiles[j].fileName,"preview.jpg") != 0)
            {
               printf("\nNot found: %s", textureFiles[j].fileName);
            }
        }
    }
    printf("\n");

    for (j = 0; j < modelFilesCount; j++)
    {
        free(modelFiles[j].fileName);
        free(modelFiles[j].originalPath);
    }
    for (j = 0; j < textureFilesCount; j++)
    {
        free(textureFiles[j].fileName);
        free(textureFiles[j].originalPath);
    }
    free(modelFiles);
    free(textureFiles);
}

int bodyPartsSizeChanged(PatcherState* patcherState)
{
    int i = 0;

    for(i = 0; i < BODY_SIZE_N_OPTIONS; i++)
    {
        if ((patcherState->bodyPartSize[i].max_size      != MAX_DEFAULT_SIZE)
        || (patcherState->bodyPartSize[i].min_size       != MIN_DEFAULT_SIZE)
        || (patcherState->bodyPartSize[i].default_size   != DEFAULT_DEFAULT_SIZE))
        {
            return 1;
        }
    }
    return 0;
}




void weaponOptionRequirePatch(FileBlock* fileBlock, PatcherState* patcherState)
{
    fileBlock->needPatch = 0;
    fileBlock->changeID = 100;



    int CLASS_MACRO = getClassMacro(fileBlock->fileName);



    if (CLASS_MACRO == NOT_FOUND)
    {
        printf("\n Failed to retrieve Class Macro  from %s in \"weaponOptionRequirePatch\"",fileBlock->fileName);
        PAUSE();
        return;
    }

    if (patcherState->weaponToRemove[CLASS_MACRO] == PRIMARY_WEAPON
        || patcherState->weaponToRemove[CLASS_MACRO] == BOTH_WEAPONS
        || patcherState->weaponToRemove[CLASS_MACRO] == REMOVE_ALL)
    {
        if (strcmpi(fileBlock->folderName,weaponsFolder[CLASS_MACRO][PRIMARY_WEAPON_FOLDER]) == 0)
        {
            if (patcherState->removeWeaponInOnly[CLASS_MACRO])
            {
                if (contains(fileBlock->fileName,"_in"))
                {
                    fileBlock->needPatch = 1;
                    fileBlock->changeID = 48;
                }
                else
                {
                     fileBlock->needPatch = 0;
                     fileBlock->changeID = 148;

                }
            }
            else
            {
                 fileBlock->needPatch = 1;
                 fileBlock->changeID = 49;
            }
        }
    }
    if (patcherState->weaponToRemove[CLASS_MACRO] == SECUNDARY_WEAPON
    || patcherState->weaponToRemove[CLASS_MACRO] == BOTH_WEAPONS
    || patcherState->weaponToRemove[CLASS_MACRO] == REMOVE_ALL)
    {
        if (strcmpi(fileBlock->folderName,weaponsFolder[CLASS_MACRO][SECUNDARY_WEAPON_FOLDER]) == 0)
        {
            if (patcherState->removeWeaponInOnly[CLASS_MACRO])
            {
                if (contains(fileBlock->fileName,"_in"))
                {
                    fileBlock->needPatch = 1;
                    fileBlock->changeID = 50;
                }
                else
                {
                     fileBlock->needPatch = 0;
                     fileBlock->changeID = 150;
                }
            }
            else
            {
                 fileBlock->needPatch = 1;
                 fileBlock->changeID = 51;
            }
        }
    }
     if (patcherState->weaponToRemove[CLASS_MACRO] == AWAKENING_WEAPON
      || patcherState->weaponToRemove[CLASS_MACRO] == REMOVE_ALL)
    {
        if (strcmpi(fileBlock->folderName,weaponsFolder[CLASS_MACRO][AWAKENING_WEAPON_FOLDER]) == 0)
        {
            if (patcherState->removeWeaponInOnly[CLASS_MACRO])
            {
                if (contains(fileBlock->fileName,"_in"))
                {
                    fileBlock->needPatch = 1;
                    fileBlock->changeID = 52;
                }
                 else
                {
                     fileBlock->needPatch = 0;
                     fileBlock->changeID = 152;
                }

            }
            else
            {
                 fileBlock->needPatch = 1;
                 fileBlock->changeID = 53;
            }
        }
    }

    // Removes Valkyrie Awakening Shield if it's set to remove
    if (CLASS_MACRO == VALKYRIE && ((patcherState->weaponToRemove[VALKYRIE] == SECUNDARY_WEAPON) || (patcherState->weaponToRemove[VALKYRIE] == REMOVE_ALL)))
    {
        if (strcmpi(fileBlock->folderName,"character/model/1_pc/7_pvw/weapon/46_spearshield/") == 0)
        {
            if (patcherState->removeWeaponInOnly[VALKYRIE])
            {
                if (contains(fileBlock->fileName,"_in"))
                {
                    fileBlock->needPatch = 1;
                    fileBlock->changeID = 54;
                }
                else
                {
                    fileBlock->needPatch = 0;
                    fileBlock->changeID = 154;
                }
            }
            else
            {

                 fileBlock->needPatch = 1;
                 fileBlock->changeID = 55;
            }
        }
    }
}

int isATreeDummerFile(char* fileName)
{
    if (contains(fileName,"treedummer"))
    {
        return 1;
    }
    if (contains(fileName,"10_hel_0004"))
    {
        return 1;
    }
     if (contains(fileName,"10_cloak_0004"))
    {
        return 1;
    }
    int i = 0;
    for (i = 0; i < TREE_DUMMER_FILES_COUNT; i++)
    {
        if (strcmpi(fileName,treedummer_files[i]) == 0)
        {
            return 1;
        }
    }
    return 0;
}


void patchFromFile(int operation, FileBlock* fileBlocks, MetaFileInfo* metaFileInfo)
{

    int skip = 0;
    int i = 0;
    int j = 0;
    int filesToPatchCount = 0;
    int paz_browser_patcher_state_n_files = 0;
    PazBrowserPatcherState* pazBrowserPatcherState =  load_paz_browser_patcher_state(&paz_browser_patcher_state_n_files, operation);

    clearFileBlocks(fileBlocks,metaFileInfo);

    if (pazBrowserPatcherState != NULL)
    {

        for (i = 0; i < metaFileInfo->fileBlocksCount; i++)
        {
            for (j = 0; j < paz_browser_patcher_state_n_files; j++)
            {
                if (strcmpi(fileBlocks[i].fileName,pazBrowserPatcherState->fileName[j]) == 0)
                {
                    fileBlocks[i].needPatch = 1;
                    filesToPatchCount++;
                    break;
                }
                if (filesToPatchCount == paz_browser_patcher_state_n_files)
                {
                    break;
                }
            }
        }
    }


    for(i = 0; i < paz_browser_patcher_state_n_files; i++)
    {
        free(pazBrowserPatcherState->fileName[i]);
    }
    free(pazBrowserPatcherState);

    if (filesToPatchCount == 0)
        {
            printf("\nWarning: No files were selected to patch in:\n");
            printf("%s\\patcher_resources\\%s", getCurrentPath(),PAZ_BROWSER_PATCHER_STATE_FILE_NAME[operation]);
            printf("\n\nCheck if the file above exists and if it has something inside of it.");
            printf("\nUse the option \"1 - Browse Files\" to mark files to patch.");
            printf("\n\nDo you want to proceed or you want to go back and change your settings first?\n");
            printf("ENTER = Proceed    ESC = Go Back\n");
            char c = '\0';

            while(c != ENTER &&  c != 10 && c != ESC)
            {
                c = getch();
            }
            if (c == ESC)
            {
                skip = 1;
            }
        }

        if (skip == 0)
        {
            if (operation == FIX_HOLES)
            {
                generatePartcutdesc(fileBlocks,metaFileInfo);
                PAUSE();
            }
            else if (operation == REMOVE_A_FILE)
            {
                printf("\n\nCopy files...\n");
                copyBlankFiles(fileBlocks,metaFileInfo);
                patchMetaFile(fileBlocks,metaFileInfo->fileBlocksCount,REMOVE_A_FILE,metaFileInfo,1);
                PAUSE();
            }
            else if (operation == RESTORE_A_FILE)
            {
                patchMetaFile(fileBlocks,metaFileInfo->fileBlocksCount,RESTORE_A_FILE,metaFileInfo,1);
                PAUSE();
            }
        }
}

